type sommet = int

type adja_graphe = (sommet*int) list array
type mat_graphe = int array array





let print_int_infty x =
  if x = Int.max_int then print_string "+∞ " else Printf.printf "%d  " x

let print_int_infty_array t =
  Array.iter (fun x -> print_int_infty x) t;
  print_newline ()

let print_int_matrix : int array array -> unit =
  Array.iter print_int_infty_array


let ex_sans_neg =
  [|
    [(1,9); (3,1)];
    [(2,5); (3,8); (4,2)];
    [(1,2)];
    [(2,3); (4,10)];
    [(2,5)]
  |]

let ex_avec_neg =
  [|
    [(1,5); (2, -4); (3,8)];
    [(0,-2)];
    [];
    [(1,-3); (2,9)];
    [(0,-6); (3,7)]
  |]


let mat_of_adja (adja : adja_graphe) = 
  let n = Array.length adja in
  let mat = Array.make_matrix n n Int.max_int in

  (** Convertit lst, la liste d'adjacence de s,
      en changeant les casses associées de mat    
  *)
  let conversion s lst =
    List.iter (fun (voisin, poids) -> mat.(s).(voisin) <- poids) lst
  in
  for s = 0 to n-1 do
    conversion s adja.(s);
    mat.(s).(s) <- 0
  done;
  mat


let copy_matrix (mat : 'a array array) =
  Array.init (Array.length mat) (fun i -> Array.copy mat.(i))

let add x y =
  if x = Int.max_int || y = Int.max_int then Int.max_int else x+y

let floyd_warshall (mat : mat_graphe) : mat_graphe =
  let n = Array.length mat in
  let dist = copy_matrix mat in

  for interm = 0 to n-1 do
    for s = 0 to n-1 do
      for t = 0 to n-1 do
        dist.(s).(t) <- min ( add dist.(s).(interm) dist.(interm).(t) ) 
                            dist.(s).(t)
      done
    done
  done;

  dist


let dijkstra (adja : adja_graphe) (s0 : sommet) : int array=
  let n = Array.length adja in
  let dist = Array.make n Int.max_int in
  let fileprio = Prio.make (n*n) 666 in

  (** Traitement appliqué à v, voisin de s,
      lors du corps de la boucle de l'algo de Dijkstra.
      p est le poids de l'arc s->v .
  *)
  let traite_un_voisin s (v,p) =
    assert (dist.(s) <> Int.max_int && p <> Int.max_int);
    if dist.(v) = Int.max_int then begin
      Prio.add fileprio v (dist.(s)+p)
    end 
    (* else ce voisin est déjà Noir *)
  in

  Prio.add fileprio s0 0;

  while not (Prio.is_empty fileprio) do
    let (u, dist_u) = Prio.take fileprio in
    if dist.(u) = Int.max_int then begin
      dist.(u) <- dist_u;
      List.iter (traite_un_voisin u) adja.(u)
    end
    (* else ce sommet est déjà Noir *)
  done;
  dist


let dist = dijkstra ex_sans_neg 0
let () = print_int_infty_array dist